package com.redhat.thermostat.plugin.eclipse.editor;

import java.net.MalformedURLException;
import java.net.URL;

import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentListener;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.forms.IManagedForm;
import org.eclipse.ui.forms.ManagedForm;
import org.eclipse.ui.forms.MasterDetailsBlock;
import org.eclipse.ui.forms.events.HyperlinkAdapter;
import org.eclipse.ui.forms.events.HyperlinkEvent;
import org.eclipse.ui.forms.widgets.FormText;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.Section;
import org.eclipse.ui.forms.widgets.TableWrapLayout;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.MultiPageEditorPart;
import org.eclipse.wst.sse.ui.StructuredTextEditor;

import com.redhat.thermostat.plugin.eclipse.model.PluginModel;

public class MultiPageEditor extends MultiPageEditorPart implements IResourceChangeListener {

    private static final int INDEX_OVERVIEW = 0;
    private static final int INDEX_COMMANDS = 1;
    private static final int INDEX_EXTENSIONS = 2;
    private static final int INDEX_SOURCE = 3;

    private StructuredTextEditor editor;
    private boolean editorIsDirty = false;

    private FormToolkit toolkit;

    private PluginModel model;
    private int oldPage = -1;

    public MultiPageEditor() {
        super();
        ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
    }

    @Override
    public void init(IEditorSite site, IEditorInput editorInput) throws PartInitException {
        if (!(editorInput instanceof IFileEditorInput))
            throw new PartInitException("Invalid Input: Must be IFileEditorInput");
        super.init(site, editorInput);

        try {
            model = new PluginModel(((IFileEditorInput)getEditorInput()).getFile());
        } catch (CoreException e) {
            model = null;
        }

    }

    @Override
    protected void createPages() {
        createOverviewPage();
        createCommandPage();
        createExtensionPage();
        createSourcePage();
    }

    private void createOverviewPage() {
        toolkit = new FormToolkit(getContainer().getDisplay());
        IManagedForm managed = new ManagedForm(getContainer());
        ScrolledForm form = managed.getForm();
        form.setText("Overview: " + model.getPluginName());
        TableWrapLayout layout = new TableWrapLayout();
        form.getBody().setLayout(layout);

        createOverviewSection(form.getBody());

        createEditCommandsAndExtensionsSection(form.getBody());

        addPage(INDEX_OVERVIEW, form);
        setPageText(INDEX_OVERVIEW, "Overview");
    }


    private void createOverviewSection(Composite parent) {
        Section section = toolkit.createSection(parent, Section.TITLE_BAR);
        section.setText("General Information");

        Composite sectionData = toolkit.createComposite(section);
        sectionData.setLayout(new TableWrapLayout());
        FormText helpText = toolkit.createFormText(sectionData, false);
        String onlineHelp = "<form>"
                + "<p>For more information about thermostat see "
                + "<a href=\"http://icedtea.classpath.org/thermostat\">the homepage</a>. "
                + "Additional information about plugins is available "
                + "<a href=\"http://icedtea.classpath.org/wiki/Thermostat/ExtensionTutorial\">here</a>."
                + "</p>"
                + "</form>";
        helpText.setText(onlineHelp, true, false);
        helpText.addHyperlinkListener(new HyperlinkAdapter() {
            @Override
            public void linkActivated(HyperlinkEvent e) {
                String href = (String) e.getHref();
                try {
                    PlatformUI.getWorkbench().getBrowserSupport().getExternalBrowser().openURL(new URL(href));
                } catch (PartInitException | MalformedURLException e1) {
                    e1.printStackTrace();
                }
            }
        });

        section.setClient(sectionData);
    }

    private void createEditCommandsAndExtensionsSection(Composite parent) {
        Section section = toolkit.createSection(parent, Section.TITLE_BAR);
        section.setText("Commands and Extensions");

        Composite sectionData = toolkit.createComposite(section);
        sectionData.setLayout(new TableWrapLayout());

        FormText commandAndExtensionsForm = toolkit.createFormText(sectionData, false);
        String formContents = ""
                + "<form>"
                + "<p>Plugins provide new commands or extend existing ones.</p>"
                + "<li><a href=\"commands\">Commands</a>: declares new commands provided by this plugin</li>"
                + "<li><a href=\"extensions\">Extensions</a>: declares which commands provided by other plugins or thermostat itself should be extended</li>"
                + "</form>";
        commandAndExtensionsForm.setText(formContents, true, false);

        commandAndExtensionsForm.addHyperlinkListener(new HyperlinkAdapter() {
            @Override
            public void linkActivated(HyperlinkEvent e) {
                String href = (String) e.getHref();
                switch(href) {
                    case "commands":
                        setActivePage(INDEX_COMMANDS);
                        break;
                    case "extensions":
                        setActivePage(INDEX_EXTENSIONS);
                        break;
                }
            }
        });

        section.setClient(sectionData);
    }

    private void createCommandPage() {
        ManagedForm managedForm = new ManagedForm(getContainer());
        ScrolledForm form = managedForm.getForm();

        form.setText("Commands");
        form.getBody().setLayout(new TableWrapLayout());

        CommandsMasterDetailsBlock masterAndDetails = new CommandsMasterDetailsBlock();
        masterAndDetails.setModel(model);
        masterAndDetails.createContent(managedForm);

        addPage(INDEX_COMMANDS, form);
        setPageText(INDEX_COMMANDS, "Commands");
    }

    private void createExtensionPage() {
        ManagedForm managedForm = new ManagedForm(getContainer());
        ScrolledForm form = managedForm.getForm();
        form.setText("Extensions");
        GridLayout layout = new GridLayout();
        layout.numColumns = 2;
        layout.makeColumnsEqualWidth = true;
        form.getBody().setLayout(layout);

        ExtensionsMasterDetailsBlock masterAndDetails = new ExtensionsMasterDetailsBlock();
        masterAndDetails.setModel(model);
        masterAndDetails.createContent(managedForm);

        addPage(INDEX_EXTENSIONS, form);
        setPageText(INDEX_EXTENSIONS, "Extensions");
    }

    void createSourcePage() {
        try {
            editor = new StructuredTextEditor();
            addPage(INDEX_SOURCE, editor, getEditorInput());
            setPageText(INDEX_SOURCE, editor.getTitle());

            getDocument().addDocumentListener(new IDocumentListener() {

                @Override
                public void documentChanged(DocumentEvent event) {
                    editorIsDirty = true;
                }

                @Override
                public void documentAboutToBeChanged(DocumentEvent event) {
                    // nothing to do
                }

            });
        } catch (PartInitException e) {
            ErrorDialog.openError(getSite().getShell(),
                    "Error creating nested text editor", null, e.getStatus());
        }
    }


    @Override
    public void dispose() {
        ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
        toolkit.dispose();
        super.dispose();
    }

    @Override
    protected void pageChange(int newPageIndex) {
        if (oldPage != -1) {
            if (oldPage == INDEX_SOURCE && editorIsDirty) {
                updateModelFromSource();
            } else if (newPageIndex == INDEX_SOURCE) {
                updateSourceFromModel();
            }
        }

        super.pageChange(newPageIndex);

        oldPage = newPageIndex;
    }

    @Override
    public void doSave(IProgressMonitor monitor) {
        if (getActivePage() != INDEX_SOURCE) {
            updateSourceFromModel();
        }

        editor.doSave(monitor);
    }

    @Override
    public boolean isSaveAsAllowed() {
        return false;
    }

    @Override
    public void doSaveAs() {
        throw new AssertionError("SaveAs is not allowed");
    }

    private void updateSourceFromModel() {
        String modelSource = model.getSourceXml();

        if (!getDocument().get().equals(modelSource)) {
            getDocument().set(modelSource);
        }

        editorIsDirty = false;
    }

    private void updateModelFromSource() {
        model.readXml(getDocument().get());

        editorIsDirty = false;
    }

    private IDocument getDocument() {
        return editor.getDocumentProvider().getDocument(editor.getEditorInput());
    }

    /**
     * Closes all project files on project close.
     */
    public void resourceChanged(final IResourceChangeEvent event) {
        if (event.getType() == IResourceChangeEvent.PRE_CLOSE) {
            Display.getDefault().asyncExec(new Runnable() {
                public void run() {
                    IWorkbenchPage[] pages = getSite().getWorkbenchWindow()
                            .getPages();
                    for (int i = 0; i < pages.length; i++) {
                        if (((FileEditorInput) editor.getEditorInput())
                                .getFile().getProject()
                                .equals(event.getResource())) {
                            IEditorPart editorPart = pages[i].findEditor(editor
                                    .getEditorInput());
                            pages[i].closeEditor(editorPart, true);
                        }
                    }
                }
            });
        }
    }

}
