package com.redhat.thermostat.plugin.eclipse.model;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Maintains a model of the plugin file.
 */
public class PluginModel {

    private String name;
    private IFile file;
    private Document doc;
    private XPathFactory xPathfactory;

    private final List<PluginModelChangeListener> listeners = new CopyOnWriteArrayList<>();

    public PluginModel(IFile file) throws CoreException {
        this(file, file.getFullPath().makeAbsolute().removeLastSegments(1).lastSegment(), file.getContents());
    }

    public PluginModel(IFile file, String name, InputStream contents) {
        this.file = file;
        this.name = name;

        this.xPathfactory = XPathFactory.newInstance();

        readXml(contents);
    }

    public void readXml(String contents) {
        ByteArrayInputStream in = new ByteArrayInputStream(contents.getBytes());
        readXml(in);
    }

    public void readXml(InputStream contents) {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder;

        InputSource inputStream = new InputSource();
        inputStream.setByteStream(contents);

        try {
            dBuilder = factory.newDocumentBuilder();
            doc = dBuilder.parse(inputStream);
        } catch (ParserConfigurationException | SAXException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IOException e) {
            throw new AssertionError("Error reading from an input stream based on a string", e);
        }

        fireModelChanged();
    }

    public String getPluginName() {
        return name;
    }

    public String getSourceXml() {
        try {
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();

            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            DOMSource source = new DOMSource(doc);
            StreamResult result = new StreamResult(outputStream);

            transformer.transform(source, result);

            return new String(outputStream.toByteArray());

            // TODO pretty print or auto-format this generated xml to look both nice and consistent
        } catch (TransformerException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return null;
    }

    public void doSaveAs() {
        // TODO Auto-generated method stub
    }

    public List<String> getCommands() {
        List<String> result = new ArrayList<>();
        XPath xpath = xPathfactory.newXPath();
        try {
            XPathExpression expr = xpath.compile("/plugin/commands/command/name/text()");
            NodeList nl = (NodeList) expr.evaluate(doc, XPathConstants.NODESET);
            for (int i = 0; i < nl.getLength(); i++) {
                Node n = nl.item(i);
                result.add(n.getNodeValue());
            }
        } catch (XPathExpressionException e) {
            e.printStackTrace();
        }
        return result;
    }

    public Command getCommand(String name) {
        Objects.requireNonNull(name);

        // FIXME sanitize name; it should be alphanumeric only!

        XPath xpath = xPathfactory.newXPath();
        String expressionString;
        XPathExpression expr;
        Node node;

        try {
            // find a command with a child 'name' having the matching text
            expressionString = "/plugin/commands/command[name/text() = '" + name + "']";
            expr = xpath.compile(expressionString);
            node = (Node) expr.evaluate(doc, XPathConstants.NODE);
            Node commandNode = node;

            expressionString = "/plugin/commands/command[name/text() = '" + name + "']/description";
            expr = xpath.compile(expressionString);
            node = (Node) expr.evaluate(doc, XPathConstants.NODE);
            String description = node == null ? "" : node.getTextContent();

            expressionString = "/plugin/commands/command[name/text() = '" + name + "']/usage";
            expr = xpath.compile(expressionString);
            node = (Node) expr.evaluate(doc, XPathConstants.NODE);
            String usage = node == null ? "" : node.getTextContent();

            List<String> arguments = parseArguments(commandNode);
            List<Option> options = parseOptions(commandNode);
            List<String> environments = parseEnvironments(commandNode);
            List<Bundle> bundles = parseBundles(commandNode);

            return new Command(name, description, usage, arguments, options, environments, bundles);
        } catch (XPathExpressionException e) {
            throw new AssertionError("bad xpath expression", e);
        }
    }

    private List<String> parseArguments(Node commandNode) throws XPathExpressionException {
        String expressionString = "arguments/argument";
        XPathExpression expr = xPathfactory.newXPath().compile(expressionString);
        NodeList nodes = (NodeList) expr.evaluate(commandNode, XPathConstants.NODESET);
        List<String> arguments = new ArrayList<>();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String arg = node.getTextContent();
            arguments.add(arg);
        }
        return arguments;
    }

    private List<Option> parseOptions(Node commandNode) {
        // TODO implement me
        return new ArrayList<>();
    }

    private List<String> parseEnvironments(Node commandNode) throws XPathExpressionException {
        String expressionString = "environments/environment";
        XPathExpression expr = xPathfactory.newXPath().compile(expressionString);
        NodeList nodes = (NodeList) expr.evaluate(commandNode, XPathConstants.NODESET);
        List<String> environments = new ArrayList<>();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            String env = node.getTextContent();
            environments.add(env);
        }
        return environments;
    }

    private List<Bundle> parseBundles(Node commandNode) throws XPathExpressionException {
        String expressionString = "bundles/bundle";
        XPathExpression expr = xPathfactory.newXPath().compile(expressionString);
        NodeList nodes = (NodeList) expr.evaluate(commandNode, XPathConstants.NODESET);
        List<Bundle> bundles = new ArrayList<>();
        for (int i = 0; i < nodes.getLength(); i++) {
            NodeList children = nodes.item(i).getChildNodes();

            String symbolicName = null;
            String version = null;

            for (int j = 0; j < children.getLength(); j++) {

                Node node = children.item(j);
                if (node.getNodeName().equals("symbolic-name")) {
                    symbolicName = node.getTextContent();
                } else if (node.getNodeName().equals("version")) {
                    version = node.getTextContent();
                }
            }

            bundles.add(new Bundle(symbolicName, version));
        }
        return bundles;
    }

    public void addCommand(String name) {
        // FIXME sanitize name; it should be alphanumeric only!

        Node root = doc;
        NodeList rootChildren = doc.getChildNodes();

        Node pluginNode = null;
        for (int i = 0; i < rootChildren.getLength(); i++) {
            Node node = rootChildren.item(i);
            if (node.getNodeName().equals("plugin")) {
                pluginNode = node;
                break;
            }
        }

        if (pluginNode == null) {
            pluginNode = doc.createElement("plugin");
            root.appendChild(pluginNode);
        }

        NodeList pluginChildren = pluginNode.getChildNodes();
        Node commandsNode = null;
        for (int i = 0; i < pluginChildren.getLength(); i++) {
            Node node = pluginChildren.item(i);
            if (node.getNodeName().equals("commands")) {
                commandsNode = node;
                break;
            }
        }

        if (commandsNode == null) {
            commandsNode = doc.createElement("commands");
            pluginNode.appendChild(commandsNode);
        }

        Node commandNode = doc.createElement("command");
        commandsNode.appendChild(commandNode);

        Node commandNameNode = doc.createElement("name");
        commandNode.appendChild(commandNameNode);

        commandNameNode.setTextContent(name);

        fireModelChanged();
    }

    public void updateCommand(Command command) {
        // FIXME sanitize name; it should be alphanumeric only!

        XPath xpath = xPathfactory.newXPath();
        String expressionString;
        XPathExpression expr;
        Node node;

        try {
            // find a command with a child 'name' having the matching text
            expressionString = "/plugin/commands/command[name/text() = '" + command.name + "']";
            expr = xpath.compile(expressionString);
            node = (Node) expr.evaluate(doc, XPathConstants.NODE);
            Node commandNode = node;

            Node descriptionNode = null;
            Node usageNode = null;
            Node argumentsNode = null;
            Node optionsNode = null;
            Node environmentsNode = null;
            Node bundlesNode = null;

            NodeList childNodes = commandNode.getChildNodes();
            for (int i = 0; i < childNodes.getLength(); i++) {
                node = childNodes.item(i);
                switch (node.getNodeName()) {
                case "description":
                    descriptionNode = node;
                    break;
                case "usage":
                    usageNode = node;
                    break;
                case "arguments":
                    argumentsNode = node;
                    break;
                case "options":
                    optionsNode = node;
                    break;
                case "environments":
                    environmentsNode = node;
                    break;
                case "bundles":
                    bundlesNode = node;
                    break;
                }
            }

            if (descriptionNode == null) {
                descriptionNode = doc.createElement("description");
                commandNode.appendChild(descriptionNode);
            }

            descriptionNode.setTextContent(command.description);

            if (usageNode == null) {
                usageNode = doc.createElement("usage");
                commandNode.appendChild(usageNode);
            }

            usageNode.setTextContent(command.usage);

            updateArguments(command.arguments, commandNode, argumentsNode);
            updateOptions(command.options, commandNode, optionsNode);
            updateEnvironments(command.environments, commandNode, environmentsNode);
            updateBundles(command.bundles, commandNode, bundlesNode);

            fireModelChanged();
        } catch (XPathExpressionException e) {
            throw new AssertionError("bad xpath expression", e);
        }
    }

    private void updateArguments(List<String> arguments, Node commandNode, Node argumentsNode) {
        if (argumentsNode == null) {
            argumentsNode = doc.createElement("arguments");
            commandNode.appendChild(argumentsNode);
        }

        // remove all existing children
        NodeList childNodes = argumentsNode.getChildNodes();
        while(childNodes.getLength() > 0) {
            Node node = childNodes.item(0);
            argumentsNode.removeChild(node);
        }

        // add new children
        for (String env : arguments) {
            Element argumentNode = doc.createElement("argument");
            argumentNode.setTextContent(env);
            argumentsNode.appendChild(argumentNode);
        }
    }

    private void updateOptions(List<Option> oiptions, Node commandNode, Node optionsNode) {
        // TODO implement me
    }

    private void updateEnvironments(List<String> environments, Node commandNode, Node environmentsNode) {
        if (environmentsNode == null) {
            environmentsNode = doc.createElement("environments");
            commandNode.appendChild(environmentsNode);
        }

        // remove all existing children
        NodeList childNodes = environmentsNode.getChildNodes();
        while(childNodes.getLength() > 0) {
            Node node = childNodes.item(0);
            environmentsNode.removeChild(node);
        }

        // add new children
        for (String env : environments) {
            Element environmentNode = doc.createElement("environment");
            environmentNode.setTextContent(env);
            environmentsNode.appendChild(environmentNode);
        }
    }

    private void updateBundles(List<Bundle> bundles, Node commandNode, Node bundlesNode) {
        if (bundlesNode == null) {
            bundlesNode = doc.createElement("bundles");
            commandNode.appendChild(bundlesNode);
        }

        // remove all existing children
        NodeList childNodes = bundlesNode.getChildNodes();
        while(childNodes.getLength() > 0) {
            Node node = childNodes.item(0);
            bundlesNode.removeChild(node);
        }

        // add new children
        for (Bundle bundle : bundles) {
            Element bundleNode = doc.createElement("bundle");

            Element nameNode = doc.createElement("symbolic-name");
            nameNode.setTextContent(bundle.name);
            Element versionNode = doc.createElement("version");
            versionNode.setTextContent(bundle.version);

            bundleNode.appendChild(nameNode);
            bundleNode.appendChild(versionNode);

            bundlesNode.appendChild(bundleNode);
        }
    }

    public void removeCommand(String name) {
        // FIXME sanitize name; it should be alphanumeric only!

        XPath xpath = xPathfactory.newXPath();

        try {
            String expressionString = "/plugin/commands/command[name/text() = '" + name + "']";
            XPathExpression expr = xpath.compile(expressionString);
            Node node = (Node) expr.evaluate(doc, XPathConstants.NODE);

            if (node == null) {
                // no match found
                return;
            }

            node.getParentNode().removeChild(node);

            fireModelChanged();
        } catch (XPathExpressionException e) {
            throw new AssertionError(e);
        }
    }

    public void addModelChangeListener(PluginModelChangeListener listener) {
        listeners.add(listener);
    }

    public void removeModelChangeListener(PluginModelChangeListener listener) {
        listeners.remove(listener);
    }

    private void fireModelChanged() {
        for (PluginModelChangeListener listener: listeners) {
            listener.modelChanged();
        }
    }

    public static class Command {
        public final String name;
        public final String description;
        public final String usage;
        public final List<String> arguments;
        public final List<Option> options;
        public final List<String> environments;
        public final List<Bundle> bundles;

        public Command(String name, String description, String usage,
                List<String> arguments, List<Option> options,
                List<String> environments, List<Bundle> bundles) {
            this.name = name;
            this.description = description;
            this.usage = usage;

            this.arguments = arguments;
            this.options = options;
            this.environments = environments;
            this.bundles = bundles;
        }
    }

    public static class Option {
        public final String name;

        public Option(String name) {
            this.name = name;
        }
    }

    public static class Extension {
        public final String name;
        public final List<Bundle> bundles;
        
        public Extension(String name){
            this.name = name;
            bundles = new ArrayList<Bundle>();
        }
        
        public Extension(String name, List<Bundle> bundles) {
            this.name = name;
            this.bundles = bundles;
        }
        
        public void addBundle(Bundle bundle){
            this.bundles.add(bundle);
        }
        
        public String getName() {
            return this.name;
        }
        
        public List<Bundle> getBundles(){
            return this.bundles;
        }
    }
    
    public static class Bundle {
        private final String name;
        private final String version;
        
        public Bundle(String name, String version) {
            this.name = name;
            this.version = version;
        }
        
        public String getName(){
            return this.name;
        }
        
        public String getVersion(){
            return this.version;
        }
    }

    public static interface PluginModelChangeListener {
        void modelChanged();
    }
    
    public List<String> getExtensions() {
        List<String> result = new ArrayList<>();
        XPath xpath = xPathfactory.newXPath();
        try {
            XPathExpression expr = xpath.compile("/plugin/extensions/extension/name/text()");
            NodeList nl = (NodeList) expr.evaluate(doc, XPathConstants.NODESET);
            for (int i = 0; i < nl.getLength(); i++) {
                Node n = nl.item(i);
                result.add(n.getNodeValue());
            }
        } catch (XPathExpressionException e) {
            e.printStackTrace();
        }
        return result;
    }
    
    public Extension getExtension(String name) {
        Objects.requireNonNull(name);

        XPath xpath = xPathfactory.newXPath();
        String expressionString;
        XPathExpression expr;
        NodeList nodeList;
        ArrayList<Bundle> bundles = new ArrayList<Bundle>();
        Bundle bundle;

        try {
            expressionString = "/plugin/extensions/extension[name/text() = '" + name + "']/bundles";
            expr = xpath.compile(expressionString);
            nodeList = (NodeList) expr.evaluate(doc, XPathConstants.NODESET);
            
            for(int i = 0; i < nodeList.getLength(); i++){
                expr = xpath.compile("//symbolic-name");
                String extensionName = (String) expr.evaluate(nodeList.item(i), XPathConstants.STRING);
                
                expr  = xpath.compile("//version");
                String extensionVersion = (String) expr.evaluate(nodeList.item(i), XPathConstants.STRING);
                
                bundle = new Bundle (extensionName, extensionVersion);
                bundles.add(bundle);
            }
        } catch (XPathExpressionException e) {
            e.printStackTrace();
        }
        return new Extension(name, bundles);
    }
    
}
